//=============================================================================
// PlungeSystem.js
//=============================================================================

//NOTE: See "Attribution" for additional licensing details. 
//      This copyright notice applies to this file (and anything
//      not in Attribution). When in doubt, just email me. :)
/**
Copyright (c) 2017 Seth N. Hetu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
**/

(function() {
  //Which plunge (in the story) is this?
  var PlungeVariableId = 2;

  //Is Risk disabled for this battle? (variable id)
  var PlungeRiskDisabledVariableId = 3;

  // Menu-aware update
  var _Scene_Map_update = Scene_Map.prototype.update;
  Scene_Map.prototype.update = function() {
    _Scene_Map_update.call(this);

    // Make our window if it doesn't exist.
    if (typeof(this._plungeControlWin) === "undefined") {
      this._plungeControlWin = new Window_PlungeControl(0, 281, 170, 178);
      this._helpWindow = new Window_PlungeHelp();
      this._helpWindow.openness = 0;
      this.addWindow(this._helpWindow);
      this._plungeControlWin.setHelpWindow(this._helpWindow);
      this._plungeControlWin.openness = 0;
      this.addWindow(this._plungeControlWin);

      //Sub-window: Attack
      this._plungeControlAttackWin = new Window_PlungeAttackControl(0, 281, 250, 178);
      this._plungeControlAttackWin.openness = 0;
      this._plungeControlAttackWin.setHelpWindow(this._helpWindow);
      this.addWindow(this._plungeControlAttackWin);

      //Sub-window: Risk
      this._plungeControlAttackRiskWin = new Window_PlungeAttackRiskControl(0, 281, 250, 178);
      this._plungeControlAttackRiskWin.openness = 0;
      this._plungeControlAttackRiskWin.setHelpWindow(this._helpWindow);
      this.addWindow(this._plungeControlAttackRiskWin);

      //Sub-window: Rinisher
      this._plungeControlFinisherWin = new Window_PlungeAttackFinisherControl(0, 281, 250, 178);
      this._plungeControlFinisherWin.openness = 0;
      this._plungeControlFinisherWin.setHelpWindow(this._helpWindow);
      this.addWindow(this._plungeControlFinisherWin);
    }

    // Are we waiting for menu input?
    if ($gameMap.isPlunging()) {
      if ($gameMap._menuMode) {
        //Open menu if needed.
        if (!this._plungeControlWin.isOpen()) {
          $gameMap.turnHasCrit = false;
          $gameMap.turnHasCritOff = false;
          $gameMap.turnHasFinishOff = 0;

          if ($gameMap.player1Left) {
            this._plungeControlWin.x = 183;
          } else {
            this._plungeControlWin.x = 462;
          }
          this._plungeControlAttackWin.x = this._plungeControlWin.x + 15;
          this._plungeControlAttackWin.y = this._plungeControlWin.y - 15;
          this._plungeControlAttackRiskWin.x = this._plungeControlAttackWin.x + 15;
          this._plungeControlAttackRiskWin.y = this._plungeControlAttackWin.y - 15;
          this._plungeControlFinisherWin.x = this._plungeControlAttackWin.x;
          this._plungeControlFinisherWin.y = this._plungeControlAttackWin.y;
          this._plungeControlAttackRiskWin.recalc();
          this._plungeControlFinisherWin.setHero($gameMap.getPlayer1());
          this._plungeControlFinisherWin.recalc();
          this._plungeControlWin._finisherReady = $gameMap.getPlayer1().finisherReady();
          this._plungeControlWin.activate();
          this._plungeControlWin.select(0);
          this._plungeControlWin.open();
          this._helpWindow.open();
        }

        //React to inputs
        if (this._plungeControlFinisherWin.isOpenAndActive()) {
          if (Input.isTriggered('ok')) {
            this.okPlungeControlFinisherWin();
          } else if (Input.isTriggered('cancel')) {
            this.cancelPlungeControlFinisherWin();
          }
        } else if (this._plungeControlAttackRiskWin.isOpenAndActive()) {
          if (Input.isTriggered('ok')) {
            this.okPlungeControlAttackRiskWin();
          } else if (Input.isTriggered('cancel')) {
            this.cancelPlungeControlAttackRiskWin();
          }
        } else if (this._plungeControlAttackWin.isOpenAndActive()) {
          if (Input.isTriggered('ok')) {
            this.okPlungeControlAttackWin();
          } else if (Input.isTriggered('cancel')) {
            this.cancelPlungeControlAttackWin();
          }
        } else if (this._plungeControlWin.isOpenAndActive()) {
          if (Input.isTriggered('ok')) {
            this.okPlungeControlWin();
          }
        }
      }
    }
  };


  //Called when the user presses something.
  Scene_Map.prototype.processPlungeTouch = function() {
    //Let the main loop catch up
    if (typeof(this._plungeControlWin) === "undefined") {
      return;
    }
    //Handle press/cancel.
    if (this._plungeControlWin.isOpen() && $gameMap._menuMode) {
      if (TouchInput.isTriggered() && this.isTouchedInsideFrame()) {
        this._touching = true;
        this.onTouch(true);
      } else if (TouchInput.isCancelled()) {
        this.processCancel();
      }
      if (this._touching) {
        if (TouchInput.isPressed()) {
          //this.onTouch(false);
        } else {
          this._touching = false;
        }
      }
    } else {
      this._touching = false;
    }
  };

  //NOTE: Do we care?
  Scene_Map.prototype.isTouchedInsideFrame = function() {
    return true;
  };


  Scene_Map.prototype.onTouch = function(triggered) {
    //Which window is active?
    var wind = undefined;
    if (this._plungeControlFinisherWin.isOpenAndActive()) {
      wind = this._plungeControlFinisherWin;
    } else if (this._plungeControlAttackRiskWin.isOpenAndActive()) {
      wind = this._plungeControlAttackRiskWin;
    } else if (this._plungeControlAttackWin.isOpenAndActive()) {
      wind = this._plungeControlAttackWin;
    } else if (this._plungeControlWin.isOpenAndActive()) {
      wind = this._plungeControlWin;
    } else {
      return;
    }

    var lastIndex = wind.index();
    var x = wind.canvasToLocalX(TouchInput.x);
    var y = wind.canvasToLocalY(TouchInput.y);
    var hitIndex = wind.hitTest(x, y);
    if (hitIndex >= 0) {
        if (hitIndex === wind.index()) {
            if (triggered /*&& wind.isTouchOkEnabled()*/) {
              //Ohhhhhh THAT's how it's supposed to work!
              if (this._plungeControlFinisherWin.isOpenAndActive()) {
                this.okPlungeControlFinisherWin();
              } else if (this._plungeControlAttackRiskWin.isOpenAndActive()) {
                this.okPlungeControlAttackRiskWin();
              } else if (this._plungeControlAttackWin.isOpenAndActive()) {
                this.okPlungeControlAttackWin();
              } else if (this._plungeControlWin.isOpenAndActive()) {
                this.okPlungeControlWin();
              }
            }
        }/* else if (wind.isCursorMovable()) {
            wind.select(hitIndex);
        }*/ //This already happens
    } /*else if (this._stayCount >= 10) {
        if (y < this.padding) {
            this.cursorUp();
        } else if (y >= this.height - this.padding) {
            this.cursorDown();
        }
    }*/ //Don't care
    /*if (this.index() !== lastIndex) {
        SoundManager.playCursor();
    }*/ //Also don't care.
  }


  Scene_Map.prototype.okPlungeControlAttackWin = function() {
    if (this._plungeControlAttackWin.isEnabled(this._plungeControlAttackWin.index())) {
      SoundManager.playOk();

      //Get the item we just picked.
      var item = this._plungeControlAttackWin.item();
      var power = item[4];

      //React to the power used.
      if (power === "normal") {
        this.plungeNormalAttack();
      } else if (power === "risk") {
        //Open the Risk sub-menu.
        this._plungeControlAttackWin.deactivate();
        this._plungeControlAttackRiskWin.activate();
        this._plungeControlAttackRiskWin.select(0);
        this._plungeControlAttackRiskWin.open();
      } else if (power === "skill") {
        this.plungeSkillAttack();
      }
    } else {
      SoundManager.playBuzzer();
    }
  }


  Scene_Map.prototype.okPlungeControlWin = function() {
    if (this._plungeControlWin.isEnabled(this._plungeControlWin.index())) {
      SoundManager.playOk();
      if (this._plungeControlWin.index() < 3) {
        this._plungeControlWin.deactivate();
        this._plungeControlAttackWin.recalc();
        this._plungeControlAttackWin.activate();
        this._plungeControlAttackWin.select(0);
        this._plungeControlAttackWin.open();
      } else {
        //Open the Finisher sub-menu.
        this._plungeControlWin.deactivate();
        this._plungeControlFinisherWin.activate();
        this._plungeControlFinisherWin.select(0);
        this._plungeControlFinisherWin.open();
      }
    } else {
      SoundManager.playBuzzer();
    }
  }


  Scene_Map.prototype.okPlungeControlFinisherWin = function() {
    if (this._plungeControlFinisherWin.isEnabled(this._plungeControlFinisherWin.index())) {
      SoundManager.playOk();

      //Get the item we just picked.
      var item = this._plungeControlFinisherWin.item();
      var power = item[4];

      //React to the power used.
      this.plungeFinisher();
    } else {
      SoundManager.playBuzzer();
    }
  }


  Scene_Map.prototype.okPlungeControlAttackRiskWin = function() {
    if (this._plungeControlAttackRiskWin.isEnabled(this._plungeControlAttackRiskWin.index())) {
      SoundManager.playOk();

      //Get the item we just picked.
      var item = this._plungeControlAttackRiskWin.item();
      var power = item[4];

      //React to the power used.
      this.plungeRiskAttack();
    } else {
      SoundManager.playBuzzer();
    }
  }


  Scene_Map.prototype.processCancel = function() {
    if (this._plungeControlFinisherWin.isOpenAndActive()) {
      this.cancelPlungeControlFinisherWin();
    } else if (this._plungeControlAttackRiskWin.isOpenAndActive()) {
      this.cancelPlungeControlAttackRiskWin();
    } else if (this._plungeControlAttackWin.isOpenAndActive()) {
      this.cancelPlungeControlAttackWin();
    }
  };


  Scene_Map.prototype.cancelPlungeControlFinisherWin = function() {
    SoundManager.playCancel();
    this._plungeControlFinisherWin.deactivate();
    this._plungeControlFinisherWin.close();
    this._plungeControlWin.activate();
  }


  Scene_Map.prototype.cancelPlungeControlAttackRiskWin = function() {
    SoundManager.playCancel();
    this._plungeControlAttackRiskWin.deactivate();
    this._plungeControlAttackRiskWin.close();
    this._plungeControlAttackWin.activate();
  }


  Scene_Map.prototype.cancelPlungeControlAttackWin = function() {
    SoundManager.playCancel();
    this._plungeControlAttackWin.deactivate();
    this._plungeControlAttackWin.close();
    this._plungeControlWin.activate();
  }


  //Called when a "Normal" attack is selected.
  Scene_Map.prototype.plungeNormalAttack = function() {
    this.selectPlungeAction();
    this.computePlungeDamage($gameMap.getPlayer1(), $gameMap.getPlayer2());
    this.computePlungeDamage($gameMap.getPlayer2(), $gameMap.getPlayer1());
    this.plungeCleanupAttack();
  };

  //Called when a "Risk" attack is selected.
  Scene_Map.prototype.plungeRiskAttack = function() {
    this.selectPlungeAction("risk");
    this.computePlungeDamage($gameMap.getPlayer1(), $gameMap.getPlayer2());
    this.computePlungeDamage($gameMap.getPlayer2(), $gameMap.getPlayer1());
    this.plungeCleanupAttack();
  }

  //Called whenever a "Finisher" is selected.
  Scene_Map.prototype.plungeFinisher = function() {
    this.selectPlungeAction("finisher");
    this.computePlungeDamage($gameMap.getPlayer1(), $gameMap.getPlayer2());
    this.computePlungeDamage($gameMap.getPlayer2(), $gameMap.getPlayer1());
    this.plungeCleanupAttack();
  }

  //Called when a Skill is selected.
  Scene_Map.prototype.plungeSkillAttack = function() {
    this.selectPlungeAction("skill");
    this.computePlungeDamage($gameMap.getPlayer1(), $gameMap.getPlayer2());
    this.computePlungeDamage($gameMap.getPlayer2(), $gameMap.getPlayer1());
    this.plungeCleanupAttack();
  }

  //Close windows, set state
  Scene_Map.prototype.plungeCleanupAttack = function() {
    $gameMap.setPlungeStateTo2();
    this._plungeControlWin.close();
    this._plungeControlAttackWin.close();
    this._helpWindow.close();
    this._plungeControlAttackRiskWin.close();
    this._plungeControlFinisherWin.close();
  }

  Scene_Map.prototype.selectPlungeAction = function(specialType) {
    //Hero action.
    var hero = $gameMap.getPlayer1();
    hero.resetPlungeAction();

    //Choose type.
    if (this._plungeControlWin.index() == 0) {
      hero._plungeAction.type = "R";
    } else if (this._plungeControlWin.index() == 1) {
      hero._plungeAction.type = "P";
    } else if (this._plungeControlWin.index() == 2) {
      hero._plungeAction.type = "S";
    }

    //Risk?
    if (specialType === "risk") {
      if (this._plungeControlAttackRiskWin.index() == 0) {
        hero._plungeAction.risk = "R";
      } else if (this._plungeControlAttackRiskWin.index() == 1) {
        hero._plungeAction.risk = "P";
      } else if (this._plungeControlAttackRiskWin.index() == 2) {
        hero._plungeAction.risk = "S";
      }
    }

    //Finisher?
    if (specialType === "finisher") {
      hero._plungeAction.finisher = true;

      //Finisher uses a different source for plunge type.
      var finishType = this._plungeControlFinisherWin.item()[4];
      hero._plungeAction.type = finishType[0];
      hero._plungeAction.finishLevel = finishType[1];
    }

    //Skill?
    if (specialType === "skill") {
      hero._plungeAction.skillId = this._plungeControlAttackWin.item()[5];
    }

    //Record stats.
    this.recordPlungeHistory(hero);

    //Foe action.
    var foe = $gameMap.getPlayer2();
    foe.resetPlungeAction();

    //Affinity is easy.
    if (foe._plungeDat.colorAffinity.length>0) {
      //Turns 0,1, 3,4, .. etc.
      if (($gameMap.plungeRound-1)%3 !== 2) {
        foe._plungeAction.type = foe._plungeDat.colorAffinity;
      } else {
        var colors = ["R","P","S"];
        var index = colors.indexOf(foe._plungeDat.colorAffinity);
        if (index >= 0) {
          colors.splice(index, 1);
        }
        foe._plungeAction.type = colors[Math.floor(Math.random()*colors.length)];
      }
    } else {
      //Just random
      var colors = ["R","P","S"];
      foe._plungeAction.type = colors[Math.floor(Math.random()*colors.length)];
      if (foe._plungeDat.willRisk) {
        this.selectRisk(hero, foe);
      }
      this.selectFinisher(hero, foe);
      this.reactToFinisher(hero, foe);
    }


    //TODO

    

    //Record stats.
    this.recordPlungeHistory(foe);
  };


  //Randomly risk
  Scene_Map.prototype.selectRisk = function(hero, foe) {
    if (Math.random() < 0.2) {
      var colors = ["R","P","S"];
      foe._plungeAction.risk = colors[Math.floor(Math.random()*colors.length)];
    }
  }

  //Use a Finisher if you have it.
  Scene_Map.prototype.selectFinisher = function(hero, foe) {
    //Any finisher ready?
    var colors = [];
    if (foe._plungeDat.tokensR>=4) {
      colors.push("R");
    }
    if (foe._plungeDat.tokensP>=4) {
      colors.push("P");
    }
    if (foe._plungeDat.tokensS>=4) {
      colors.push("S");
    }

    if (colors.length>0) {
      foe._plungeAction.type = colors[Math.floor(Math.random()*colors.length)];
      foe._plungeAction.risk = "";
      foe._plungeAction.finisher = true;
    }
  }


  //Try to avoid being trounced by a finisher.
  Scene_Map.prototype.reactToFinisher = function(hero, foe) {
    //Special case when the hero has a Finisher ready.
    if (foe._plungeDat.finisherBlock) {
      foe._plungeDat.finisherBlock = false;
    } else {
      //Any finisher ready?
      var colors = [];
      if (hero._plungeDat.tokensR>=4) {
        colors.push("P");
      }
      if (hero._plungeDat.tokensP>=4) {
        colors.push("S");
      }
      if (hero._plungeDat.tokensS>=4) {
        colors.push("R");
      }
      if (colors.length>0) {
        foe._plungeAction.type = colors[Math.floor(Math.random()*colors.length)];
        foe._plungeDat.finisherBlock = true;
      }
    }
  }


  //Save relevant stats about this decision
  Scene_Map.prototype.recordPlungeHistory = function(chara) {
    //Who is this?
    var name = $gameActors.actor(chara._plungeId).name();
    var act = chara._plungeAction;
    var key = "general";

    //TODO: History is a little bit busted; we can "risk" on a different color; right now it's just a bool

    //Always set general
    $gameParty.setHistoryStat(name, key, act.type, act.risk, false);

    //TurnX
    key = "turn" + $gameMap.plungeRound;
    $gameParty.setHistoryStat(name, key, act.type, act.risk, false);


    //TODO: Log more stuff!
  }

  // "source" of the plunge action (i.e., the hero)
  // "tagett" of the plunge action (i.e., the foe)
  // Note that some things affect the source too (like healing)
  Scene_Map.prototype.computePlungeDamage = function(source, target) {
    //Roll damage
    var baseDmg = 0;
    for (var i=0; i<source._plungeDat.currPlunge; i++) {
      baseDmg = (Math.random() < 0.5 ? 1 : 2);
    }

    //RPS Bonus
    var rpsPlus = 0;
    if (this.isStrong(source._plungeAction.type, target._plungeAction.type)) {
      rpsPlus = Math.round(0.5 * baseDmg);
      if (rpsPlus <= 0) {
        rpsPlus = 1;
      }
    } else if (this.isWeak(source._plungeAction.type, target._plungeAction.type)) {
      rpsPlus = -1 * Math.round(0.5 * baseDmg);
      if (rpsPlus >= 0) {
        rpsPlus = -1;
      }
    }

    //Risk bonus/detrement
    var riskPlus = 0;
    if (source._plungeAction.risk.length > 0) {
      if (this.isStrong(source._plungeAction.risk, target._plungeAction.type)) {
        riskPlus = Math.round(0.25 * baseDmg);
      } else if (this.isWeak(source._plungeAction.risk, target._plungeAction.type)) {
        riskPlus = -1 * Math.round(0.25 * baseDmg);

        //No tokens won.
        source._plungeAction.risk = "";
      }
    }

    //Crit.
    var critPlus = 0;
    if (Math.random()*100 <= source._plungeDat.currCritPercent) {
      critPlus = Math.round((baseDmg*source._plungeDat.currCritDamage)/100);
      if (critPlus <= 0) {
        $gameMap.turnHasCrit = true;
        $gameMap.turnHasCritOff = true;
        source._plungeAction.critical = true;
        critPlus = 1;
      }
    }

    //Finisher mult AFTER everything else.
    var finishAdd = 0;
    var finishMult = 1.0;
    if (source._plungeAction.finisher) {
      $gameMap.turnHasFinishOff = 3;
      if (source._plungeAction.finishLevel == 5) {
        finishMult = 1.8;
        finishAdd = 10;
      } else if (source._plungeAction.finishLevel == 4) {
        finishMult = 1.6;
        finishAdd = 7;
      } else if (source._plungeAction.finishLevel == 3) {
        finishMult = 1.5;
        finishAdd = 4;
      } else if (source._plungeAction.finishLevel == 2) {
        finishMult = 1.4;
        finishAdd = 2;
      } else {
        finishMult = 1.25;
        finishAdd = 2;
      }

      //Some enemies are very weak to finishers
      if (target._plungeDat.isFrail) {
        finishMult *= 1.5;
        finishAdd *= 2;
      }
    }

    //Set damage received by target
    var newAmt = Math.round(finishMult * (baseDmg + rpsPlus + riskPlus + critPlus)) + finishAdd;
    if (newAmt <= 0) {
      newAmt = 1;
    }
    target._plungeAction.damage += newAmt;

    //Jagged edge
    if (source._plungeAction.skillId === 2) {
      target._plungeAction.damage += 2 + Math.floor(Math.random() * 3);
      $gameMap.getPlayer1()._plungeDat.currFp = Math.max(0, $gameMap.getPlayer1()._plungeDat.currFp - $dataSkills[2].mpCost);
    }

    //Source non-damage effects (healing, skill-up)
    if (source._plungeAction.skillId === 1) {
      source._plungeAction.healing = 2 + Math.floor(Math.random() * 3);
      $gameMap.getPlayer1()._plungeDat.currFp = Math.max(0, $gameMap.getPlayer1()._plungeDat.currFp - $dataSkills[1].mpCost);
    }

    //Target non-damage effects (skill-down)
    //TODO

    //Injury?
    var dmg = target._plungeAction.damage;
    var dt = target._plungeDat.currDT;
    if (dmg > dt+10+1.1*dt) {
      target._plungeAction.injury = 4;
    } else if (dmg > dt+5+0.8*dt) {
      target._plungeAction.injury = 3;
    } else if (dmg > dt+2+0.4*dt) {
      target._plungeAction.injury = 2;
    } else if (dmg > dt+1) {
      target._plungeAction.injury = 1;
    }

    //Special case
    if (source._plungeAction.finishLevel>=4) {
      target._plungeAction.injury = 4;
    }

    //Tokens the source received for this action.
    //NOTE: We need to figure out how the hero accumulates gray tokens from the point of view of the foe.
    if (!source._plungeAction.finisher) {
      source._plungeAction.tokens.push(source._plungeAction.type);
      if (source._plungeAction.risk.length>0) {
        if (source._plungeId === 2) {
          var colors = ["R","P","S"];
          for (var i=0; i<colors.length; i++) {
            var clr = colors[i];
            if (!this.isWeak(clr, target._plungeAction.type)) {
              source._plungeAction.tokens.push(clr+"i"); //TODO: "i" is currently busted.
            }
          }
        } else {
          source._plungeAction.tokens.push(source._plungeAction.risk);
        }
      }
    }
  }

  //Is type1 strong/Weak against type2?
  //TODO: Finishers and other have special consideration.
  Scene_Map.prototype.isStrong = function(type1, type2) {
    if (type1.toLowerCase() === "r") {
      return type2.toLowerCase() === "s";
    } else if (type1.toLowerCase() === "p") {
      return type2.toLowerCase() === "r";
    } else if (type1.toLowerCase() === "s") {
      return type2.toLowerCase() === "p";
    } else if (type1.toLowerCase() === "u") {
      return true;
    }
    return false;
  }
  Scene_Map.prototype.isWeak = function(type1, type2) {
    if (type1.toLowerCase() === "u") {
      return false;
    } else if (type1.toLowerCase() === "r") {
      return type2.toLowerCase() === "p";
    } else if (type1.toLowerCase() === "p") {
      return type2.toLowerCase() === "s";
    } else if (type1.toLowerCase() === "s") {
      return type2.toLowerCase() === "r";
    }
    return false;
  }


  // Only show the equip menu.
  // TODO: Later, abilities? (Or always have them on?)
  Scene_Map.prototype.callMenu = function() {
    SoundManager.playOk();
    SceneManager.push(Scene_Equip);
    Window_MenuCommand.initCommandPosition();
    $gameTemp.clearDestination();
    this._mapNameWindow.hide();
    this._waitCount = 2;
  };

  // Modify our equipment menu to show HP/MP
  var _Window_EquipStatus_drawItem = Window_EquipStatus.prototype.drawItem;
  Window_EquipStatus.prototype.drawItem = function(x, y, paramId) {
    //Start at HP/MP
    _Window_EquipStatus_drawItem.call(this, x, y, paramId-2);
  };

  // Make sure we always heal the player to full after changing equipment
  var _Scene_Equip_onItemOk = Scene_Equip.prototype.onItemOk;
  Scene_Equip.prototype.onItemOk = function() {
    _Scene_Equip_onItemOk.call(this);
    var actors = $gameParty.members();
    for (var i=0; i<actors.length; i++) {
      actors[i].recoverAll();
    }

    //...and save
    DataManager.saveGame(1);
  }

  //////////////////////////////////
  // Plunge control selector (main)
  //////////////////////////////////

  function Window_PlungeControl() {
    this.initialize.apply(this, arguments);
  };

  Window_PlungeControl.prototype = Object.create(Window_Selectable.prototype);
  Window_PlungeControl.prototype.constructor = Window_PlungeControl;

  Window_PlungeControl.prototype.initialize = function(x, y, width, height) {
    Window_Selectable.prototype.initialize.call(this, x, y, width, height);
    this._itemsList = [
      "Raging Slash",
      "Poised Cut",
      "Serene Sweep",
      "Finisher"
    ];
    this.finisherColor = $FinishColor;
    this.finisherTimer = 0;
    this.contents.fontSize = 20;
    this.refresh();
  };

  Window_PlungeControl.prototype.standardBackOpacity = function() {
    return 255;
  };
  Window_PlungeControl.prototype.loadWindowskin = function() {
    this.windowskin = ImageManager.loadSystem('Window2');
  };

  Window_PlungeControl.prototype.update = function() {
    Window_Selectable.prototype.update.call(this);

    //Animate finisher
    if (this._finisherReady) {
      this.finisherTimer += 1;
      this.undoFinishColor = true;
      var oldColor = this.finisherColor;
      if (this.finisherTimer >= 60) {
        this.finisherTimer = 0;
        this.finisherColor = $FinishColor;
      } else if (this.finisherTimer >= 50) {
        this.finisherColor = $SerenityColor;
      } else if (this.finisherTimer >= 40) {
        this.finisherColor = $FinishColor;
      } else if (this.finisherTimer >= 30) {
        this.finisherColor = $PoiseColor;
      } else if (this.finisherTimer >= 20) {
        this.finisherColor = $FinishColor;
      } else if (this.finisherTimer >= 10) {
        this.finisherColor = $RageColor;
      }

      if (oldColor !== this.finisherColor) {
        this.clearItem(3);
        this.drawItem(3);
      }
    } else if (this.undoFinishColor) {
      this.clearItem(3);
      this.drawItem(3);
      this.undoFinishColor = false;
    }
  };

  Window_PlungeControl.prototype.maxItems = function() {
    return 4;
  };

  Window_PlungeControl.prototype.item = function() {
    return this._itemsList[this.index()];
  };

  Window_PlungeControl.prototype.indexedColor = function(index) {
    if (index == 0) {
      return $RageColor;
    } else if (index == 1) {
      return $PoiseColor;
    } else if (index == 2) {
      return $SerenityColor;
    } else if (index == 3) {
      if (this.isEnabled(index)) {
        return this.finisherColor;
      } else {
        return "#999999";
      }
    }
  }

  Window_PlungeControl.prototype.drawItem = function(index) {
    var rect = this.itemRectForText(index);
    this.changeTextColor(this.indexedColor(index));
    this.drawText(this._itemsList[index], rect.x, rect.y, 138, this.lineHeight());
    this.changePaintOpacity(true);
  };

  Window_PlungeControl.prototype.isEnabled = function(index) {
      return index<3 || this._finisherReady;
  };

  Window_PlungeControl.prototype.isCurrentItemEnabled = function() {
      return this.isEnabled(this.index());
  };

  Window_PlungeControl.prototype.itemHeight = function() {
    return this.lineHeight() - 2;
  };

  // TODO
  Window_PlungeControl.prototype.updateHelp = function() {
    Window_Selectable.prototype.updateHelp.call(this);
    var txt = "";
    if (this.index() == 0) {
      txt = "Strong against Serence Sweep. Weak against Poised Cut.";
    } else if (this.index() == 1) {
      txt = "Strong against Raging Slash. Weak against Serence Sweep.";
    } else if (this.index() == 2) {
      txt = "Strong against Poised Cust. Weak against Raging Slash.";
    } else if (this.index() == 3) {
      txt = "A powerful finishing move, weak against nothing.";
      if (!this.isEnabled(this.index())) {
        txt += " (Not ready yet!)";
      }
    }
    this._helpWindow.setText(txt);
  };


  ////////////////////////////////
  // Secondary selection window after choosing one of the 3 attacks
  ////////////////////////////////

  function Window_PlungeAttackControl() {
    this.initialize.apply(this, arguments);
  };

  Window_PlungeAttackControl.prototype = Object.create(Window_Selectable.prototype);
  Window_PlungeAttackControl.prototype.constructor = Window_PlungeAttackControl;

  Window_PlungeAttackControl.prototype.initialize = function(x, y, width, height) {
    Window_Selectable.prototype.initialize.call(this, x, y, width, height);
    this.recalc();
    this.contents.fontSize = 20;
  };

  Window_PlungeAttackControl.prototype.standardBackOpacity = function() {
    return 255;
  };
  Window_PlungeAttackControl.prototype.loadWindowskin = function() {
    this.windowskin = ImageManager.loadSystem('Window2');
  };


  Window_PlungeAttackControl.prototype.recalc = function() {
    //Rebuild the item list. Note:
    // displayName, displayColor, helpText, isEnabled, powerWord, (args)
    this._itemsList = [
      ["Normal", "#FFFFFF", "Perform a Normal attack of this type", true, "normal"]  //We can always do a normal attack.
    ];

    //We usually list Risk attacks too.
    if ($gameVariables.value(PlungeRiskDisabledVariableId) === 0) {
      this._itemsList.push (
        ["Risk", "#FFFF00", "Perform a Risky attack of this type", true, "risk"]
      );
    } else {
      this._itemsList.push (
        ["Risk", "#999999", "Perform a Risky attack of this type (Temporarily Disabled)", false, "risk"]
      );
    }

    //After that come Blade Arts.
    if (SceneManager._scene._plungeControlWin.index() == 2) {
      if ( $gameActors.actor(1).isLearnedSkill(1)) {
        var enabled = $dataSkills[1].mpCost <= $gameMap.getPlayer1()._plungeDat.currFp;
        this._itemsList.push (
          ["Cool Wind", (enabled ? "#00FFFF" : "#999999"), "Heal 2 to 4 HP after attacking", enabled, "skill", 1]
        );
      }
    }
    if (SceneManager._scene._plungeControlWin.index() == 0) {
      if ( $gameActors.actor(1).isLearnedSkill(2)) {
        var enabled = $dataSkills[2].mpCost <= $gameMap.getPlayer1()._plungeDat.currFp;
        this._itemsList.push (
          ["Jagged Edge", (enabled ? "#FF6633" : "#999999"), "Deal 2 to 4 extra damage", enabled, "skill", 2]
        );
      }
    }

    this.refresh();
  }

  Window_PlungeAttackControl.prototype.update = function() {
    Window_Selectable.prototype.update.call(this);
  };

  Window_PlungeAttackControl.prototype.maxItems = function() {
    if (typeof(this._itemsList) === "undefined") {
      return 0;
    }
    return this._itemsList.length;
  };

  Window_PlungeAttackControl.prototype.item = function() {
    return this._itemsList[this.index()];
  };

  Window_PlungeAttackControl.prototype.drawItem = function(index) {
    var rect = this.itemRectForText(index);
    this.changeTextColor(this._itemsList[index][1]);
    var xOff = 0;
    if (this._itemsList[index][4] === "skill") {
      this.drawIcon($dataSkills[this._itemsList[index][5]].iconIndex, rect.x, rect.y+1);
      rect.x += 16+21;
    }
    this.drawText(this._itemsList[index][0], rect.x+xOff, rect.y, 138, this.lineHeight());
    this.changePaintOpacity(true);
  };

  Window_PlungeAttackControl.prototype.isEnabled = function(index) {
      return this._itemsList[index][3];
  };

  Window_PlungeAttackControl.prototype.isCurrentItemEnabled = function() {
      return this.isEnabled(this.index());
  };

  Window_PlungeAttackControl.prototype.itemHeight = function() {
    return this.lineHeight() - 2;
  };

  Window_PlungeAttackControl.prototype.updateHelp = function() {
    Window_Selectable.prototype.updateHelp.call(this);
    if (this.index() >= 0 && this.index()<this._itemsList.length) {
      var txt = this._itemsList[this.index()][2];
      this._helpWindow.setText(txt);
    }
  };


  ////////////////////////////////////
  // Select which Risk attack to use
  ////////////////////////////////////

  function Window_PlungeAttackRiskControl() {
    this.initialize.apply(this, arguments);
  };

  Window_PlungeAttackRiskControl.prototype = Object.create(Window_Selectable.prototype);
  Window_PlungeAttackRiskControl.prototype.constructor = Window_PlungeAttackRiskControl;

  Window_PlungeAttackRiskControl.prototype.initialize = function(x, y, width, height) {
    Window_Selectable.prototype.initialize.call(this, x, y, width, height);
    this.recalc();
    this.contents.fontSize = 20;
  };

  Window_PlungeAttackRiskControl.prototype.standardBackOpacity = function() {
    return 255;
  };
  Window_PlungeAttackRiskControl.prototype.loadWindowskin = function() {
    this.windowskin = ImageManager.loadSystem('Window2');
  };


  Window_PlungeAttackRiskControl.prototype.recalc = function() {
    //This one's easy.
    this._itemsList = [
      ["Risk Rage", $RageColor, "Risk a Rage attack for a token and extra damage", true, "riskRage"],
      ["Risk Poise", $PoiseColor, "Risk a Poise attack for a token and extra damage", true, "riskPoise"],
      ["Risk Serene", $SerenityColor, "Risk a Serene attack for a token and extra damage", true, "riskSerenity"],
    ];

    this.refresh();
  }

  Window_PlungeAttackRiskControl.prototype.update = function() {
    Window_Selectable.prototype.update.call(this);
  };

  Window_PlungeAttackRiskControl.prototype.maxItems = function() {
    if (typeof(this._itemsList) === "undefined") {
      return 0;
    }
    return this._itemsList.length;
  };

  Window_PlungeAttackRiskControl.prototype.item = function() {
    return this._itemsList[this.index()];
  };

  Window_PlungeAttackRiskControl.prototype.drawItem = function(index) {
    var rect = this.itemRectForText(index);
    this.changeTextColor(this._itemsList[index][1]);
    this.drawText(this._itemsList[index][0], rect.x, rect.y, 138, this.lineHeight());
    this.changePaintOpacity(true);
  };

  Window_PlungeAttackRiskControl.prototype.isEnabled = function(index) {
      return this._itemsList[index][3];
  };

  Window_PlungeAttackRiskControl.prototype.isCurrentItemEnabled = function() {
      return this.isEnabled(this.index());
  };

  Window_PlungeAttackRiskControl.prototype.itemHeight = function() {
    return this.lineHeight() - 2;
  };

  Window_PlungeAttackRiskControl.prototype.updateHelp = function() {
    Window_Selectable.prototype.updateHelp.call(this);
    if (this.index() >= 0) {
      var txt = this._itemsList[this.index()][2];
      this._helpWindow.setText(txt);
    }
  };


  /////////////////////////////////////////
  // Finisher selection
  /////////////////////////////////////////

  function Window_PlungeAttackFinisherControl() {
    this.initialize.apply(this, arguments);
  };

  Window_PlungeAttackFinisherControl.prototype = Object.create(Window_Selectable.prototype);
  Window_PlungeAttackFinisherControl.prototype.constructor = Window_PlungeAttackFinisherControl;

  Window_PlungeAttackFinisherControl.prototype.initialize = function(x, y, width, height) {
    Window_Selectable.prototype.initialize.call(this, x, y, width, height);
    this.recalc();
    this.contents.fontSize = 20;
  };

  Window_PlungeAttackFinisherControl.prototype.standardBackOpacity = function() {
    return 255;
  };
  Window_PlungeAttackFinisherControl.prototype.loadWindowskin = function() {
    this.windowskin = ImageManager.loadSystem('Window2');
  };

  Window_PlungeAttackFinisherControl.prototype.setHero = function(hero) {
    this._hero = hero;
  }

  Window_PlungeAttackFinisherControl.prototype.recalc = function() {
    if (typeof(this._hero) === "undefined") { return; }

    //Tokens available determines the level.
    var dat = this._hero._plungeDat;
    var levelR = Math.min(4, Math.floor(dat.tokensR / 4));
    var levelP = Math.min(4, Math.floor(dat.tokensP / 4));
    var levelS = Math.min(4, Math.floor(dat.tokensS / 4));
    var levelUlt = (levelR==4 && levelP==4 && levelS==4) ? 5 : 0;

    this._itemsList = [];
    if (levelUlt>0) {
      this._itemsList.push(
        ["Ultimate Finisher", "#FFFF00", "Ultimate multi-form Finisher", true, ["U",levelUlt]]
      );
    }
    if (levelR>0) {
      this._itemsList.push(
        ["Rage Finisher "+levelR, $RageColor, "Level " + levelR + " Rage Finisher", true, ["R",levelR]]
      );
    }
    if (levelP>0) {
      this._itemsList.push(
        ["Poise Finisher "+levelP, $PoiseColor, "Level " + levelP + " Poise Finisher", true, ["P",levelP]]
      );
    }
    if (levelS>0) {
      this._itemsList.push(
        ["Serene Finisher "+levelS, $SerenityColor, "Level " + levelS + " Serene Finisher", true, ["S",levelS]]
      );
    }

    this.refresh();
  }

  Window_PlungeAttackFinisherControl.prototype.update = function() {
    Window_Selectable.prototype.update.call(this);
  };

  Window_PlungeAttackFinisherControl.prototype.maxItems = function() {
    if (typeof(this._itemsList) === "undefined") {
      return 0;
    }
    return this._itemsList.length;
  };

  Window_PlungeAttackFinisherControl.prototype.item = function() {
    return this._itemsList[this.index()];
  };

  Window_PlungeAttackFinisherControl.prototype.drawItem = function(index) {
    var rect = this.itemRectForText(index);
    this.changeTextColor(this._itemsList[index][1]);
    this.drawText(this._itemsList[index][0], rect.x, rect.y, 138, this.lineHeight());
    this.changePaintOpacity(true);
  };

  Window_PlungeAttackFinisherControl.prototype.isEnabled = function(index) {
      return this._itemsList[index][3];
  };

  Window_PlungeAttackFinisherControl.prototype.isCurrentItemEnabled = function() {
      return this.isEnabled(this.index());
  };

  Window_PlungeAttackFinisherControl.prototype.itemHeight = function() {
    return this.lineHeight() - 2;
  };

  Window_PlungeAttackFinisherControl.prototype.updateHelp = function() {
    Window_Selectable.prototype.updateHelp.call(this);
    if (this.index() >= 0) {
      if (this.index() < this._itemsList.length) {
        var txt = this._itemsList[this.index()][2];
        this._helpWindow.setText(txt);
      }
    }
  };


  ////////////////////////////////
  // Smaller help window
  ////////////////////////////////

  function Window_PlungeHelp() {
    this.initialize.apply(this, arguments);
  }

  Window_PlungeHelp.prototype = Object.create(Window_Help.prototype);
  Window_PlungeHelp.prototype.constructor = Window_PlungeHelp;

  Window_PlungeHelp.prototype.initialize = function(numLines) {
    var buffX = 5;
    var buffY = 5;
    var width = Graphics.boxWidth - 2*buffX;
    var height = this.fittingHeight(1);
    Window_Base.prototype.initialize.call(this, buffX, buffY, width, height);
    this._text = '';

    this.contents.fontSize = 22;
    this.contents.textColor = "#FFFFFF";
  };

  Window_PlungeHelp.prototype.standardBackOpacity = function() {
    return 255;
  };
  Window_PlungeHelp.prototype.loadWindowskin = function() {
    this.windowskin = ImageManager.loadSystem('Window2');
  };

  Window_PlungeHelp.prototype.refresh = function() {
    this.contents.clear();
    this.drawText(this._text, 0, 0, this.contents.rect.width, "left");
  };

  //Smaller text.
  Window_PlungeHelp.prototype.fittingHeight = function(numLines) {
    return Window_Base.prototype.fittingHeight.call(this, numLines) - 3;
  };



})();
