//=============================================================================
// PlungeSystem.js
//=============================================================================

//NOTE: See "Attribution" for additional licensing details. 
//      This copyright notice applies to this file (and anything
//      not in Attribution). When in doubt, just email me. :)
/**
Copyright (c) 2017 Seth N. Hetu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
**/

//Colors for our major properties.
var $RageColor = "#FF6666";
var $PoiseColor = "#66CC66";
var $SerenityColor = "#0099FF";
var $FinishColor = "#FFFF00";

//For easy debugging
var $ForceLosePlunge = false;
var $ForceWinPlunge = false;

(function() {
  //Switch for:
  // 0 => not in a plunge
  // 1 => please init the plunge
  // 2 => in a plunge
  // 3 => won the plunge
  // 4 => lost the plunge (or tied)
  var PlungeVariableId = 1;

  //Set this to pause menu execution (useful for events)
  var PlungeMenuHoldSwitch = 10;

  //Switches used to call back to events in the map
  var PlungeTurn1Switch = 1; //Turn 1 just started.
  var PlungeTurn2Switch = 2; //Turn 2 just started.
  var PlungeTurn3Switch = 3; //Turn 3 just started.
  var PlungeTurn4Switch = 4; //Turn 4 just started.
  var PlungeTurn5Switch = 5; //Turn 5 just started.

  //How long to wait when showing standard damage, nothing, etc.
  var PlungeNoneCountdown = 50;
  var PlungeStandardCountdown = 70;
  var PlungeCollapseTimer = 50;

  //Variables controlling our hero's plunge stats.
  var HeroMaxHPVar = 10;
  var HeroMaxMPVar = 11;
  var HeroMaxPlungeVr = 12;
  var HeroMaxDmgThreshVar = 13;
  var HeroMaxCritPercVar = 14;
  var HeroMaxCritDmgVar = 15;

  //Variables controlling the current enemy's plunge stats.
  var FoeMaxHPVar = 21;
  var FoeMaxMPVar = 22;
  var FoeMaxPlungeVr = 23;
  var FoeMaxDmgThreshVar = 24;
  var FoeMaxCritPercVar = 25;
  var FoeMaxCritDmgVar = 26;

  //More foe traits
  var FoeAffinity = 27; //Color (1=R, 2=P, 3=S, 0=None) that dictates action. If >0, two turns in a row of that color, then randomly choose another, then repeat.
  var FoeIsFrail = 28;  //If >0, foe is frail (double-damage from finishers)
  var FoeUsesRisk = 29; //If>0, this foe uses Risk attacks sometimes

  StorageManager.deleteSlot = function(savefileId) {
    try {
      StorageManager.remove(savefileId);
    } catch (e2) {
    }
  }

  // Skip title screen.
  Scene_GameEnd.prototype.commandToTitle = function() {
    this.fadeOutAll();
    this.startNewGameOrLoadOngoing();
  };
  Scene_Gameover.prototype.gotoTitle = function() {
    this.startNewGameOrLoadOngoing();
  };
  Scene_Boot.prototype.start = function() {
    Scene_Base.prototype.start.call(this);
    SoundManager.preloadImportantSounds();
    if (DataManager.isBattleTest()) {
      DataManager.setupBattleTest();
      SceneManager.goto(Scene_Battle);
    } else if (DataManager.isEventTest()) {
      DataManager.setupEventTest();
      SceneManager.goto(Scene_Map); //Don't care if this works.
    } else {
      this.checkPlayerLocation();
      this.startNewGameOrLoadOngoing();
      Window_TitleCommand.initCommandPosition();
    }
    this.updateDocumentTitle();
  };

  Scene_Base.prototype.setupNewVars = function() {
    //Set our inital plunge values.
    //NOTE: These are automatically overwritten, but it's good enough for now.
    $gameVariables.setValue(HeroMaxHPVar, 20);
    $gameVariables.setValue(HeroMaxMPVar, 10);
    $gameVariables.setValue(HeroMaxPlungeVr, 2);
    $gameVariables.setValue(HeroMaxDmgThreshVar, 4);
    $gameVariables.setValue(HeroMaxCritPercVar, 10);
    $gameVariables.setValue(HeroMaxCritDmgVar, 20);

    //Force our hero to be transparent for the opening cutscene.
    if ($dataSystem.startMapId === 2) {
      var actor = $gameActors.actor(1);
      if (actor) {
        actor.setCharacterImage("Empty", 0);
      }
    }

    //Initialize our stats structure.
    // "name" => {stats} //name = "hero", "john", etc., based on actor 1/2
    // stats are:
    //   "somestat" => [r, p, s, riskR, riskP, riskS, finisher]
    // ...where "r,p,s" are "times an attack has been selected for 'somestat' condition"
    // ...and "riskR/P/S/finish" are "times a risk attack of that color (or any finisher) were selected for 'somestat' condition"
    // ...r/p/s = total times that condition occurred (excluding finishers), riskR+P+S+finish = total number of risk attacks and finishers for that condition.
    // including:
    //   "general" => Any time it was that person's turn.
    //   "turnX" => Decision on turn X (1==first) in any battle.
    //   "highHP" => Whenever HP at the start of that turn is green+
    //   "medHP" => Whenever HP at the start of that turn is yellow to green
    //   "lowHP" => Whenever HP at the start of that turn is red to yellow
    $gameParty.trackStats = {}
  };

  //Get/Set sanely
  Game_Party.prototype.getHistoryStat = function(name, key) {
    if (typeof(this.trackStats[name]) === "undefined") {
      this.trackStats[name] = {};
    }
    if (typeof(this.trackStats[name][key]) === "undefined") {
      this.trackStats[name][key] = [0,0,0, 0,0,0,0];
    }
    return this.trackStats[name][key];
  };
  Game_Party.prototype.setHistoryStat = function(name, key, rps, isRisk, isFinish) {
    var val = this.getHistoryStat(name, key);

    var subKey = 99;

    //Finish/RPS
    if (isFinish) {
      subKey = 6;
    } else if (rps.toLowerCase() === "r") {
      subKey = 0;
    } else if (rps.toLowerCase() === "p") {
      subKey = 1;
    } else if (rps.toLowerCase() === "s") {
      subKey = 2;
    }

    //Risk?
    if (subKey>=0 && subKey<=2 && isRisk) {
      subKey += 3;
    }

    if (subKey>=0 && subKey<=6) {
      val[subKey] += 1;
    }
  };

  // Load a save file if it exists.
  Scene_Base.prototype.startNewGameOrLoadOngoing = function() {
    if (DataManager.loadGame(1)) {
      console.log("Loading auto-sve");
      if ($gameSystem.versionId() !== $dataSystem.versionId) {
        $gamePlayer.reserveTransfer($gameMap.mapId(), $gamePlayer.x, $gamePlayer.y);
        $gamePlayer.requestMapReload();
      }
    } else {
      console.log("Starting new game");
      DataManager.setupNewGame();
      this.setupNewVars();
    }
    SceneManager.goto(Scene_Map);
  }

  // Make sure we have our palettes loaded.
  var _Spriteset_Map_createCharacters = Spriteset_Map.prototype.createCharacters;
  Spriteset_Map.prototype.createCharacters = function() {
    _Spriteset_Map_createCharacters.call(this);

    //Get our number bitmap.
    if (typeof($numberBitmap) === "undefined") {
      $numberBitmap = ImageManager.loadSystem('Number');
    }
  }

  // Load plunge properties at start.
  var _Game_Map_setupEvents = Game_Map.prototype.setupEvents;
  Game_Map.prototype.setupEvents = function() {
    _Game_Map_setupEvents.call(this);

    this.plunge = false;
    this.computePlunge();
    if (this.plunge) {
      this.computePlungeEvents();
    }
  };

  var _Game_Map_update = Game_Map.prototype.update;
  Game_Map.prototype.update = function(sceneActive) {
    _Game_Map_update.call(this, sceneActive);

    if (this.wantsToPlunge()) {
      this.initPlunge();
      $gameVariables.setValue(PlungeVariableId, 2);
    }

    if (this.isPlunging()) {
      this.updatePlunge();
    }
  };

  //Is this a plunge map? Identifies the <plunge> tag.
  Game_Map.prototype.computePlunge = function() {
    //TODO: We can hook this into Yanfly's generic notetag processor.
    var noteReg = /<(?:PLUNGE)>/i;
    var notedata = $dataMap.note.split(/[\r\n]+/);

    for (var i = 0; i < notedata.length; i++) {
      var line = notedata[i];
      if (line.match(noteReg)) {
        console.log("Plunge is ON");
        this.plunge = true;
      }
    }
  };

  //Perform initialization, once per plunge.
  Game_Map.prototype.initPlunge = function() {
    //This controls the entire plunge state machine:
    // 1 = idle
    // 2 = running fast towards the center.
    // 3 = jumping with sword out towards the top-center
    // 4 = (attack anim?)
    // 5 = falling down away from the center.
    // 6 = running fast away from the center.
    // =>back to idle
    // 10 = falling down (victory/defeat).
    // 11 = falling down anim over, end plunge.
    this.plungeState = 1;

    //Damage waiting to be shown. 
    //(Will be shown in the next phase 1)
    //Contains arrays of the form:
    //  [target(gameEvent), type(string), amount(int), ... (depends on type)]
    this.pendingDamage = [];
    this.currDamage = null; //What are we currently showing?
    
    //What around are we on? 
    this.plungeRound = 1;

    //Which side are we on?
    this.player1Left = this._events[this.player1]._realX < this._events[this.player2]._realX;

    //Set plunge stats for each player.
    this._events[this.player1].initPlunge();
    this._events[this.player2].initPlunge();

    //Allow turn 1 switch
    this._turnJustStarted = true;
  }

  //Should be called when the plunge is done.
  Game_Map.prototype.plungeDone = function() {
    this._events[this.player1]._characterName = this._events[this.player1]._origCharacterName;
    this._events[this.player2]._characterName = this._events[this.player2]._origCharacterName;
  }

  //Grab the players participating in the plunge and save them.
  Game_Map.prototype.computePlungeEvents = function() {
    var noteReg = /^<player([12])>$/i;
    var zoomReg = /^<ZoomPoint>$/i;

    this.player1 = null;
    this.player2 = null;
    this.zoomPt = null;
    for (var i = 0; i < $dataMap.events.length; i++) {
      if ($dataMap.events[i]) {
        var ev = $dataMap.events[i]
        if (ev.note.match(noteReg)) {
          console.log(`Plunge Player ${RegExp.$1} is: ${ev.name}`);
          if (RegExp.$1 == 1) {
            this.player1 = i;
          } else {
            this.player2 = i;
          }
        } else if (ev.note.match(zoomReg)) {
          this.zoomPt = i;
        }
      }
    }

    //It's an error to make a plunge map without both participants.
    if ((this.player1 == null) || (this.player2 == null)) {
      throw `Can't find both players: ${this.player1} , ${this.player2}`;
    }
    if (this.zoomPt == null) {
      throw `Can't find zoom point.`;
    }

    //Save IDs
    this._events[this.player1]._plungeId = 1;
    this._events[this.player2]._plungeId = 2;
  };

  //Disable movement/menu when plunging.
  var _Game_Player_update = Game_Player.prototype.update;
  Game_Player.prototype.update = function(sceneActive) {
    if (!$gameMap.isPlunging()) {
      _Game_Player_update.call(this, sceneActive);
    }
  }
  var _Scene_Map_processMapTouch = Scene_Map.prototype.processMapTouch;
  Scene_Map.prototype.processMapTouch = function() {
    if ($gameMap.isPlunging()) {
      this.processPlungeTouch();
    } else {
      _Scene_Map_processMapTouch.call(this);
    }
  }
  var _Scene_Map_updateCallMenu = Scene_Map.prototype.updateCallMenu;
  Scene_Map.prototype.updateCallMenu = function() {
    if (!$gameMap.isPlunging()) {
      _Scene_Map_updateCallMenu.call(this);
    }
  }


  Game_Character.prototype.initPlunge = function() {
    //Whose stats are we loading?
    var varHP = FoeMaxHPVar;
    var varMP = FoeMaxMPVar;
    var varPL = FoeMaxPlungeVr;
    var varDT = FoeMaxDmgThreshVar;
    var varCP = FoeMaxCritPercVar;
    var varCD = FoeMaxCritDmgVar;
    if (this._plungeId === 1) {
      varHP = HeroMaxHPVar;
      varMP = HeroMaxMPVar;
      varPL = HeroMaxPlungeVr;
      varDT = HeroMaxDmgThreshVar;
      varCP = HeroMaxCritPercVar;
      varCD = HeroMaxCritDmgVar;
    }

    //NOTE: For the hero, we need to actually pull these from storage.
    if (this._plungeId === 1) {
      var actor = $gameActors.actor(1);
      if (actor) {
        $gameVariables.setValue(HeroMaxHPVar, actor.hp);
        $gameVariables.setValue(HeroMaxMPVar, actor.mp);
        $gameVariables.setValue(HeroMaxPlungeVr, actor.param(2));
        $gameVariables.setValue(HeroMaxDmgThreshVar, actor.param(3));
        $gameVariables.setValue(HeroMaxCritPercVar, actor.param(4));
        $gameVariables.setValue(HeroMaxCritDmgVar, actor.param(5));
      }
    }

    //Basic plunge data
    this._plungeDat = {};
    this._plungeDat.maxHp = $gameVariables.value(varHP);
    this._plungeDat.currHp = this._plungeDat.maxHp;
    this._plungeDat.maxFp = $gameVariables.value(varMP);
    this._plungeDat.currFp = this._plungeDat.maxFp;
    this._plungeDat.maxDT = $gameVariables.value(varDT);
    this._plungeDat.currDT = this._plungeDat.maxDT;
    this._plungeDat.maxPlunge = $gameVariables.value(varPL);
    this._plungeDat.currPlunge = this._plungeDat.maxPlunge;
    this._plungeDat.maxCritPercent = $gameVariables.value(varCP);
    this._plungeDat.currCritPercent = this._plungeDat.maxCritPercent;
    this._plungeDat.maxCritDamage = $gameVariables.value(varCD);
    this._plungeDat.currCritDamage = this._plungeDat.maxCritDamage;

    //How many tokens we are known to have
    this._plungeDat.tokensR = 0;
    this._plungeDat.tokensP = 0;
    this._plungeDat.tokensS = 0;

    //How many tokens we might have. (Usually only applies to enemies).
    this._plungeDat.tokensPossibleR = 0;
    this._plungeDat.tokensPossibleP = 0;
    this._plungeDat.tokensPossibleS = 0;

    //Have we tried blocking a Finisher?
    this._plungeDat.finisherBlock = false;

    //Does this Plunger have a color affinity? (R, P, S, or "")
    if (this._plungeId === 1) {
      this._plungeDat.colorAffinity = "";
    } else {
      this._plungeDat.colorAffinity = "";
      if ($gameVariables.value(FoeAffinity) === 1) {
        this._plungeDat.colorAffinity = "R";
      } else if ($gameVariables.value(FoeAffinity) === 2) {
        this._plungeDat.colorAffinity = "P";
      } else if ($gameVariables.value(FoeAffinity) === 3) {
        this._plungeDat.colorAffinity = "S";
      }
    }

    //Is this Plunger weak to Finishers?
    if (this._plungeId === 1) {
      this._plungeDat.isFrail = false;
    } else {
      this._plungeDat.isFrail = $gameVariables.value(FoeIsFrail)>0;
    }

    //Will this Plunger use Risk attacks?
    if (this._plungeId === 1) {
      this._plungeDat.willRisk = false;
    } else {
      this._plungeDat.willRisk = $gameVariables.value(FoeUsesRisk)>0;
    }

    console.log(this._plungeDat);

    //Might as well reset this too.
    this.resetPlungeAction();
  }

  //Do we have at least enough tokens for one finisher?
  Game_Character.prototype.finisherReady = function() {
    return this._plungeDat.tokensR>=4 || this._plungeDat.tokensP>=4 || this._plungeDat.tokensS>=4;
  };

  Game_Character.prototype.resetPlungeAction = function() {
    //Data relating to the current attack.
    this._plungeAction = {};
    this._plungeAction.type = "R";  //R,P,S
    this._plungeAction.risk = "";   //R,P,S, or "" for "no risk"
    this._plungeAction.skillId = 0; //What skill are we using?
    this._plungeAction.critical = false; //Was this a critical action?
    this._plungeAction.finisher = false; //Was this a finisher?
    this._plungeAction.finishLevel = 0;  //Strength of finisher attack.
    this._plungeAction.damage = 0;  //Damage YOU'VE taken from this+opponent's plunge action.
    this._plungeAction.healing = 0; //Health YOU'VE gained from this+oponent's plunge action.
    this._plungeAction.injury = 0; //Injury level sustained for this round.
    this._plungeAction.tokens = []; //Tokens (R,P,S) YOU'VE gained from this+opponent's plunge action. "i" after a token means "imaginary"; i.e., unknown.
  }

  // Actually apply the plunge action to our stats.
  Game_Character.prototype.applyPlungeAction = function() {
    this._plungeDat.currHp -= this._plungeAction.damage;
    if (this._plungeDat.currHp > 0) {
      this._plungeDat.currHp += this._plungeAction.healing;
    }

    //Apply DT.
    if (this._plungeAction.injury == 1) {
      this._plungeDat.currDT = Math.max(0, this._plungeDat.currDT-1);
    } else if (this._plungeAction.injury == 2) {
      this._plungeDat.currDT = Math.max(0, this._plungeDat.currDT-2);
      this._plungeDat.currPlunge = Math.max(0, this._plungeDat.currPlunge-1);
    } else if (this._plungeAction.injury == 3) {
      this._plungeDat.currDT = Math.max(0, this._plungeDat.currDT-3);
      this._plungeDat.currPlunge = Math.max(0, this._plungeDat.currPlunge-2);
    } else if (this._plungeAction.injury == 4) {
      this._plungeDat.currHp = 0;
    }

    //Apply tokens unless we used a finisher (which clears them)
    if (this._plungeAction.finisher) {
      this._plungeDat.tokensR = 0;
      this._plungeDat.tokensP = 0;
      this._plungeDat.tokensS = 0;
      this._plungeDat.tokensPossibleR = 0;
      this._plungeDat.tokensPossibleP = 0;
      this._plungeDat.tokensPossibleS = 0;
    } else {
      for (var i=0; i<this._plungeAction.tokens.length; i++) {
        var token = this._plungeAction.tokens[i];
        if (token.startsWith("R")) {
          if (token.endsWith("i")) {
            this._plungeDat.tokensPossibleR += 1;
          } else {
            this._plungeDat.tokensR += 1;
          }
        } else if (token.startsWith("P")) {
          if (token.endsWith("i")) {
            this._plungeDat.tokensPossibleP += 1;
          } else {
            this._plungeDat.tokensP += 1;
          }
        } else if (token.startsWith("S")) {
          if (token.endsWith("i")) {
            this._plungeDat.tokensPossibleS += 1;
          } else {
            this._plungeDat.tokensS += 1;
          }
        }
      }
    }
  }

  // Queue up the anims we show directly after landing.
  Game_Character.prototype.queueStatUpdate = function() {
    $gameMap.pendingDamage.push([this, "None", PlungeNoneCountdown]);
    if (this._plungeAction.injury>0) {
      $gameMap.pendingDamage.push([this, "Injury", this._plungeAction.damage]);
    } else {
      $gameMap.pendingDamage.push([this, "Standard", this._plungeAction.damage]);
    }
    if (this._plungeDat.currHp > 0) {
      if (this._plungeAction.healing > 0) {
        $gameMap.pendingDamage.push([this, "None", PlungeNoneCountdown/2]);
        $gameMap.pendingDamage.push([this, "Standard", -this._plungeAction.healing]);
      }
      if (this._plungeAction.tokens.length > 0) {
        $gameMap.pendingDamage.push([this, "None", PlungeNoneCountdown/2]);
        $gameMap.pendingDamage.push([this, "Tokens", this._plungeAction.tokens]);
      }
    }
  };

  Game_Map.prototype.isPlunging = function() {
    return $gameVariables.value(PlungeVariableId) === 2;
  }

  Game_Map.prototype.wantsToPlunge = function() {
    return $gameVariables.value(PlungeVariableId) === 1;
  }

  //Check victory for hero or enemy.
  Game_Map.prototype.plungeVictory = function() {
    //Special case
    if ($ForceWinPlunge && $gameMap.plungeRound>=3) {
      return true;
    }

    return this._events[this.player2]._plungeDat.currHp <= 0;
  }
  Game_Map.prototype.plungeDefeat = function() {
    //Special case
    if ($ForceLosePlunge && $gameMap.plungeRound>=2) {
      return true;
    }

    return this._events[this.player1]._plungeDat.currHp <= 0;
  }

  //Helpers for Scene_Map
  Game_Map.prototype.getPlayer1 = function() {
    return this._events[this.player1];
  }
  Game_Map.prototype.getPlayer2 = function() {
    return this._events[this.player2];
  }

  Game_Map.prototype.setPlungeStateTo2 = function() {
    var players = [null, this.player1, this.player2];
    this.plungeState = 2;
    for (var i=1; i<=2; i++) {
      this._events[players[i]]._plungePattern[0] = "Swing";
      this._events[players[i]]._startTile = [this._events[players[i]]._realX,this._events[players[i]]._realY];
    }

    for (var i=1; i<=2; i++) {
      this._events[players[i]]._turnOffLifeCycle = true;
    }
  }

  Game_Map.prototype.updatePlunge = function() {
    var players = [null, this.player1, this.player2];

    //Update phase.
    this._menuMode = false;
    if (this.plungeState == 1) {
      if (this.currDamage !== null) {
        // Countdown
        this.currCountdown -= 1;
        if (this.currCountdown <= 0) {
          this.currDamage[0].updateDamage(null);
          this.currDamage = null;
        } else {
          //Move?
          if (this.currDamage[1] === "Standard" || this.currDamage[1] === "Injury" || this.currDamage[1] === "Tokens") {
            if (this.currCountdown >= PlungeStandardCountdown-10) {
              this.currDamage[0]._dmgOffsetY -= 3;
            }
          }
        }
      } else if (this.pendingDamage.length>0) {
        this.currDamage = this.pendingDamage.shift();
        //Show the initial damage value.
        this.currCountdown = 0;
        if (this.currDamage[1] === "Standard" || this.currDamage[1] === "Injury") {
          this.currDamage[0]._dmgOffsetY = 20;
          this.currDamage[0].updateDamage(this.currDamage[2], this.currDamage[1] === "Injury");
          this.currCountdown = PlungeStandardCountdown; 
        } else if (this.currDamage[1] === "Tokens") {
          this.currDamage[0]._dmgOffsetY = 20;
          this.currDamage[0].updateDamageTokens(this.currDamage[2]);
          this.currCountdown = PlungeStandardCountdown; 
        } else if (this.currDamage[1] === "None") {
          this.currDamage[0].updateDamage(null);
          this.currCountdown = this.currDamage[2];
        } //TODO: Other damage types (Injury, etc.)
      } else if (this.plungeVictory() || this.plungeDefeat()) {
        //Start collapse
        this._events[this.player1].collapseTimer = 0;
        this._events[this.player2].collapseTimer = 0;
        if (this.plungeVictory()) {
          this._events[this.player2]._plungePattern[1] = 1;
          this._events[this.player2].collapseTimer = PlungeCollapseTimer;
          this._events[this.player2].collapseState = 0;
        }
        if (this.plungeDefeat()) {
          this._events[this.player1]._plungePattern[1] = 1;
          this._events[this.player1].collapseTimer = PlungeCollapseTimer;
          this._events[this.player1].collapseState = 0;
        }
        this.plungeState = 10;
      } else {
        for (var i=1; i<=2; i++) {
          this._events[players[i]]._turnOnLifeCycle = true;
        }

        //Turn this on the first time.
        if (this._turnJustStarted) {
          this._turnJustStarted = false;
          if (this.plungeRound==1) {
            $gameSwitches.setValue(PlungeTurn1Switch, true);
          } else if (this.plungeRound==2) {
            $gameSwitches.setValue(PlungeTurn2Switch, true);
          } else if (this.plungeRound==3) {
            $gameSwitches.setValue(PlungeTurn3Switch, true);
          } else if (this.plungeRound==4) {
            $gameSwitches.setValue(PlungeTurn4Switch, true);
          } else if (this.plungeRound==5) {
            $gameSwitches.setValue(PlungeTurn5Switch, true);
          }
        } else {

          //Can't use the menu while an event is running.
          if (!$gameSwitches.value(PlungeMenuHoldSwitch)) {
            this._menuMode = true;
          }
        }
      }
    } else if (this.plungeState == 10) {
      //Update death anim.
      for (var i=1; i<=2; i++) {
        var ev = this._events[players[i]];
        if (ev.collapseTimer > 0) {
          ev.collapseTimer -= 1;
          if (ev.collapseTimer <= 0) {
            ev.collapseState += 1;
            if (ev.collapseState === 1) {
              ev._plungePattern[0] = "ReadyMagical";
            } else if (ev.collapseState === 2) {
              ev._plungePattern[0] = "Crisis";
            } else if (ev.collapseState === 3) {
              ev._plungePattern[0] = "Dead";
            } else if (ev.collapseState === 4) {
              this.plungeState = 11;
            }
            ev.collapseTimer = PlungeCollapseTimer;
          }
        }
      }
    } else if (this.plungeState == 11) {
      if (this.plungeDefeat()) {
        $gameVariables.setValue(PlungeVariableId, 4);
      } else if (this.plungeVictory()) {
        $gameVariables.setValue(PlungeVariableId, 3);
      } else {
        $gameVariables.setValue(PlungeVariableId, 0); //This should never happen, but just for safety
      }
      return;
    }

    //Update movement.
    var phaseOver = false;
    if ($gameMap.plungeState == 2) {
      //Run 3 tiles.
      for (var i=1; i<=2; i++) {
        var sign = ($gameMap.player1Left ? (i==1) : (i==2)) ? 1 : -1;
        this._events[players[i]]._realX += sign * 0.1;
        if (Math.abs(this._events[players[i]]._realX - this._events[players[i]]._startTile[0]) >= 3) {
          phaseOver = true;
          this._events[players[i]]._plungePattern[0] = "Item";
          this._events[players[i]]._plungePattern[1] = 1;
          this._events[players[i]].slashAnim = false;
          this._events[players[i]].landAnim = false;
        }
      }
    } else if ($gameMap.plungeState == 3) {
      //Zoom in (fast) for crits
      if ($gameMap.turnHasCrit) {
        $gameMap.setZoom(2.5, 2.5, 25);
        $gameMap.setZoomCenter(this._events[this.zoomPt]);
        $gameMap.turnHasCrit = false;
      }

      //Wait for zoom
      if (Math.abs($gameMap._destZoom.x-$gameMap._zoom.x)<=0.001) {
        //Jump 3 tiles.
        for (var i=1; i<=2; i++) {
          var sign = ($gameMap.player1Left ? (i==1) : (i==2)) ? 1 : -1;
          this._events[players[i]]._realX += sign * 0.1;
          this._events[players[i]]._realY -= 0.125;
          if (Math.abs(this._events[players[i]]._realX - this._events[players[i]]._startTile[0]) >= 5.5) {
            //Start anim slightly early
            if (!this._events[players[i]].slashAnim) {
              var mirror = ($gameMap.player1Left ? (i==1) : (i==2));
              var otherId = i==1 ? 2 : 1;
              this._events[players[i]].requestAnimation([this._events[players[otherId]].slashAnimRPS(),mirror]);
              this._events[players[i]].slashAnim = true;
            }
          }
          if (Math.abs(this._events[players[i]]._realX - this._events[players[i]]._startTile[0]) >= 6) {
            phaseOver = true;
            this._events[players[i]]._plungePattern[0] = "Evade";
            this._events[players[i]]._plungePattern[1] = 1;
          }
        }
      }
    } else if ($gameMap.plungeState == 4) {
      //Empty phase (maybe wait?)
      phaseOver = true;
    } else if ($gameMap.plungeState == 5) {
      //Zoom out (fast) for crits
      if (!(this._events[this.player1].isAnimationPlaying() || this._events[this.player2].isAnimationPlaying())) {
        if ($gameMap.turnHasCritOff) {
          $gameMap.setZoom(1, 1, 15);
          //$gameMap.setZoomCenter(this._events[this.zoomPt]);
          $gameMap.turnHasCritOff = false;
        }
        if ($gameMap.turnHasFinishOff > 0) {
          $gameMap.turnHasFinishOff -= 1;
        }
      }

      //Wait for zoom
      if (!$gameMap.turnHasCritOff && $gameMap.turnHasFinishOff<=0 && Math.abs($gameMap._destZoom.x-$gameMap._zoom.x)<=0.001) {
        //Fall 3 tiles.
        for (var i=1; i<=2; i++) {
          var sign = ($gameMap.player1Left ? (i==1) : (i==2)) ? 1 : -1;
          this._events[players[i]]._realX += sign * 0.1;
          this._events[players[i]]._realY += 0.125;

          //Show a healing, etc. anim?
          if (Math.abs(this._events[players[i]]._realX - this._events[players[i]]._startTile[0]) >= 8.5) {
            if (!this._events[players[i]].landAnim) {
              //Show a healing anim?
              var mirror = ($gameMap.player1Left ? (i==1) : (i==2));
              if (this._events[players[i]]._plungeAction.healing > 0) {
                this._events[players[i]].requestAnimation([45,mirror]);
              } else if (this._events[this.player1]._plungeAction.skillId === 2) {
                this._events[this.player2].requestAnimation([130,!mirror]);
              }
              this._events[players[i]].landAnim = true;
            }
          }

          if (Math.abs(this._events[players[i]]._realX - this._events[players[i]]._startTile[0]) >= 9) {
            phaseOver = true;
            this._events[players[i]]._plungePattern[0] = "Swing";
            this._events[players[i]]._plungePattern[1] = 1;
          }
        }
      }
    } else if ($gameMap.plungeState == 6) {
      //Run 3 tiles.
      for (var i=1; i<=2; i++) {
        var sign = ($gameMap.player1Left ? (i==1) : (i==2)) ? 1 : -1;
        this._events[players[i]]._realX += sign * 0.1;
        if (Math.abs(this._events[players[i]]._realX - this._events[players[i]]._startTile[0]) >= 12) {
          phaseOver = true;
          this._events[players[i]]._plungePattern[0] = "Idle";
          this._events[players[i]]._plungePattern[1] = 1;
        }
      }
    }

    //Update
    if (phaseOver) {
      $gameMap.plungeState += 1;
      if ($gameMap.plungeState >= 7) {
        //Turn over
        $gameMap.plungeState = 1;
        $gameMap.player1Left = !$gameMap.player1Left;
        this.plungeRound += 1;

        //Flip charas
        for (var i=1; i<=2; i++) {
          this._events[players[i]]._startTile = undefined;
          if ($gameMap.player1Left) {
            this._events[players[i]].flipH = i==1;
          } else {
            this._events[players[i]].flipH = i==2;
          }
        }

        //Apply our plunge actions.
        for (var i=1; i<=2; i++) {
          this._events[players[i]].applyPlungeAction();
        }

        //Queue up our damage/healing/token anims
        for (var i=1; i<=2; i++) {
          this._events[players[i]].queueStatUpdate();
        }

        //At this point, we don't need any more plunge state.
        for (var i=1; i<=2; i++) {
          this._events[players[i]].resetPlungeAction();
        }

        //Allow turn X switch
        this._turnJustStarted = true;
      }
    }


    //Update anim.
    for (var i=1; i<=2; i++) {
      //Change graphics, if appropriate.
      var nm = "#" + $gameActors.actor(parseInt(i)).battlerName();
      if (this._events[players[i]].characterName() !== nm) {
        this._events[players[i]]._origCharacterName = this._events[players[i]]._characterName;
        this._events[players[i]].forceBattleChara(nm);
        if ($gameMap.player1Left) {
          this._events[players[i]].flipH = i==1;
        } else {
          this._events[players[i]].flipH = i==2;
        }
      }

      //Step
      this._events[players[i]].forceStepAnim();
    }
  }

  Game_Character.prototype.slashAnimRPS = function() {
    if (this._plungeAction.finisher) {
      if (this._plungeAction.type.toLowerCase() === "r") {
        return 11;
      } else if (this._plungeAction.type.toLowerCase() === "p") {
        return 12;
      } else if (this._plungeAction.type.toLowerCase() === "s") {
        return 13;
      }

      return 10; //Ultimate Finisher
    } else if (this._plungeAction.critical) {
      if (this._plungeAction.type.toLowerCase() === "r") {
        return 7;
      } else if (this._plungeAction.type.toLowerCase() === "p") {
        return 8;
      } else if (this._plungeAction.type.toLowerCase() === "s") {
        return 9;
      }

      return 6; //Safety
    } else {
      if (this._plungeAction.type.toLowerCase() === "r") {
        return 3;
      } else if (this._plungeAction.type.toLowerCase() === "p") {
        return 4;
      } else if (this._plungeAction.type.toLowerCase() === "s") {
        return 5;
      }
      return 2; //Safety
    }
  };

  //Allow mirrored chara anims
  Sprite_Character.prototype.setupAnimation = function() {
    var animId = this._character.animationId();
    var mirror = false;
    if (animId && animId.constructor === Array) {
      mirror = animId[1];
      animId = animId[0];
    }

    if (animId > 0) {
        var animation = $dataAnimations[animId];
        this.startAnimation(animation, mirror, 0);
        this._character.startAnimation();
    }
  };

  Game_CharacterBase.prototype.forceBattleChara = function(name) {
    this._characterName = name;
    this._plungePattern = ["Idle",1];  //"Idle", center
  }

  Game_CharacterBase.prototype.plungePattern = function() {
    return this._plungePattern;
  };

  Game_CharacterBase.prototype.forceStepAnim = function() {
    var animWait = this.animationWait();

    //Faster animation when running
    if ($gameMap.plungeState == 2 || $gameMap.plungeState == 5) {
      animWait /= 2;
    }

    //No anim when jumping with sword out or dying.
    if ($gameMap.plungeState == 3 || $gameMap.plungeState == 10) {
      return;
    }

    //Update the plunge anim (0, 1, 2, 3, 0), where 3 == 1
    if (typeof(this._plungePattern) !== "undefined") {
      this._animationCount++;
      if (this._animationCount >= animWait) {
        this._animationCount = 0;
        this._plungePattern[1] = (this._plungePattern[1] + 1) % 4;
      }
    }
  }

  //Don't mess with our moving charas.
  var _Game_CharacterBase_updateMove = Game_CharacterBase.prototype.updateMove;
  Game_CharacterBase.prototype.updateMove = function() {
    if (typeof(this._plungePattern) === "undefined") {
      _Game_CharacterBase_updateMove.call(this);
    }
  }


  //Redraw and show lifecycle.
  Sprite_Character.prototype.turnOnLifeCycle = function() {
    //Position
    var onLeft = $gameMap.player1Left ? this._character._plungeId==1 : this._character._plungeId==2;
    if (onLeft) {
      this._lifeCycle.x = 12;
    } else {
      this._lifeCycle.x = 464;
    }
    this._lifeCycle.y = 475;

    //State
    var bmp = this._lifeCycle.bitmap;
    var dat = this._character._plungeDat;
    bmp.fillRect(0,0,bmp.rect.width,bmp.rect.height, "#FFFFFF");
    bmp.fillRect(3,3,bmp.rect.width-6,bmp.rect.height-6, "#000000");

    //HP/FP/DT/PL
    bmp.textColor = "#FFFFFF";

    //Name
    var sX = 10;
    var sY = 10;
    var halfX = bmp.rect.width/2+10-3;
    bmp.fontSize = 24;
    var name = $gameActors.actor(this._character._plungeId).name();
    bmp.drawText(name, sX, sY, 100, 21, "left");

    //Stats
    sX = 10;
    sY += 25;
    bmp.fontSize = 18;
    bmp.drawText("HP:", sX, sY, 100, 21, "left");
    this.drawHealthBar(bmp, sX+30, sY, 120, 12, dat.currHp, dat.maxHp);
    sX = halfX;
    bmp.drawText("FP:", sX, sY, 100, 21, "left");
    this.drawHealthBar(bmp, sX+30, sY, 120, 12, dat.currFp, dat.maxFp);
    sX = 10;
    sY += 25;
    bmp.drawText("DT:", sX, sY, 100, 21, "left");
    this.drawHealthBar(bmp, sX+30, sY, 120, 12, dat.currDT, dat.maxDT);
    sX = halfX;
    bmp.drawText("PL:", sX, sY, 100, 21, "left");
    this.drawHealthBar(bmp, sX+30, sY, 120, 12, dat.currPlunge, dat.maxPlunge);

    //Hide unknown tokens for player.
    var tokensPossibleR = dat.tokensPossibleR;
    var tokensPossibleP = dat.tokensPossibleP;
    var tokensPossibleS = dat.tokensPossibleS;
    if (this._character._plungeId==1) {
      tokensPossibleR = 0;
      tokensPossibleP = 0;
      tokensPossibleS = 0;
    }

    //RPS gauge
    sY += 25;
    sX = 10;
    bmp.fontSize = 14;
    bmp.drawText("Rage:", sX, sY, 50, 21, "right");
    this.drawRPSbar(bmp, sX+50, sY, $RageColor, dat.tokensR, tokensPossibleR);
    sY += 15;
    bmp.drawText("Poise:", sX, sY, 50, 21, "right");
    this.drawRPSbar(bmp, sX+50, sY, $PoiseColor, dat.tokensP, tokensPossibleP);
    sY += 15;
    bmp.drawText("Serenity:", sX, sY, 50, 21, "right");
    this.drawRPSbar(bmp, sX+50, sY, $SerenityColor, dat.tokensS, tokensPossibleS);

    //Opacity 
    this._lifeCycle.opacity = 255;
  }


  Sprite_Character.prototype.drawHealthBar = function(bmp, sx, sy, sw, sh, currVal, maxVal) {
    //Background
    sy += 5;
    bmp.fillRect(sx, sy, sw, sh, "#FFFFFF");
    sx += 1; sy += 1; sw -= 2; sh -= 2;
    bmp.fillRect(sx, sy, sw, sh, "#000000");
    sx += 1; sy += 1; sw -= 2; sh -= 2;
    bmp.fillRect(sx, sy, sw, sh, "#222222");

    //Fill color
    fract = (currVal*100)/maxVal;
    if (fract > 100) {
      color = "#4169E1";
    } else if (fract >= 80) {
      color = "#228B22";
    } else if (fract >= 60) {
      color = "#9ACD32";
    } else if (fract >= 40) {
      color = "#FFFF00";
    } else if (fract >= 20) {
      color = "#FFA500";
    } else {
      color = " #FF0000";
    }

    //Fill bar length
    nw = Math.round(sw * fract) / 100;
    if (nw < sw && nw >= 0) {
      sw = nw;
    }
    bmp.fillRect(sx, sy, sw, sh, color);
  }

  Sprite_Character.prototype.drawRPSbar = function(bmp, sx, sy, haveColor, tokens, tokensPossible) {
    sx += 8;
    sy += 11;
    for (var i=1; i<=20; i++) {
      //Finish early?
      if (i > tokens + tokensPossible) { break; }

      //Draw the circle
      bmp.drawCircle(sx, sy, 4, "#FFFFFF");
      if (i <= tokens) {
        bmp.drawCircle(sx, sy, 3, haveColor);
      } else if (i <= tokens + tokensPossible) {
        bmp.drawCircle(sx, sy, 3, "#333333");

        //Color blind
        bmp.fillRect(sx-2, sy-2, 1, 1, "#CCCCCC");
        bmp.fillRect(sx-1, sy-1, 1, 1, "#CCCCCC");
        bmp.fillRect(sx+1, sy+1, 1, 1, "#CCCCCC");
        bmp.fillRect(sx, sy, 1, 1, "#CCCCCC");
      }
      sx += 12;

      //Draw the seperator?
      if (i==4 || i==8 || i == 12 || i == 16) {
        bmp.fillRect(sx-3, sy-7, 1, 15, "#FFFFFF");
        sx += 6;
      }
    }
  }


  //Properly set the sprite's bitmap.
  var _Sprite_Character_setCharacterBitmap = Sprite_Character.prototype.setCharacterBitmap;
  Sprite_Character.prototype.setCharacterBitmap = function() {
    //Reset if plunge is done.
    if (!$gameMap.isPlunging() && this._characterName.startsWith("#")) {
      this._characterName = this._characterName.substring(1);
    }

    if (this._characterName.startsWith("#")) {
      // TEMP: Do this now.
      if (typeof(this._damageSprites) === "undefined") {
        //Make our damage sprites.
        this._damageSprites = [new Sprite(), new Sprite(), new Sprite()];
        for (var i=0; i<3; i++) {
          var spr = this._damageSprites[i];
          spr._dmgOffsetX = 0;
          spr.bitmap = new Bitmap($numberBitmap.width/10, $numberBitmap.height/4);
          spr.opacity = 0;
          SceneManager._scene._spriteset.addChild(spr);
        }
      }
      if (typeof(this._lifeCycle) === "undefined") {
        //Make our damage sprites.
        this._lifeCycle = new Sprite();
        this._lifeCycle.bitmap = new Bitmap(340, 140);
        this._lifeCycle.opacity = 0;
        SceneManager._scene._spriteset.addChild(this._lifeCycle);
      }

      // Back to our regularly scheduled functionality.
      this.bitmap = ImageManager.loadSvActor(this._characterName.substring(1));
      this._isBigCharacter = false;

      //Flip?
      if (this._character.flipH) {
        this.scale.x = -1 * Math.abs(this.scale.x);
      } else {
        this.scale.x = 1 * Math.abs(this.scale.x);
      }
    } else {
      //Default
      _Sprite_Character_setCharacterBitmap.call(this);
    }
  };


  //For the next round
  Game_Event.prototype.updateDamage = function(amt, isInjury) {
    this._newDamageRequest = amt;
    this._newDamageIsInjury = isInjury;
  }
  Game_Event.prototype.updateDamageTokens = function(tokens) {
    this._newDamageRequestTokens = tokens;
  }

  Game_CharacterBase.prototype.getNewDamageRequest = function() {
    var res = this._newDamageRequest;
    this._newDamageRequest = undefined;
    return res;
  }
  Game_CharacterBase.prototype.getNewDamageInjury  = function() {
    var res = this._newDamageIsInjury;
    this._newDamageIsInjury = undefined;
    return res;
  }
  Game_CharacterBase.prototype.getNewDamageRequestTokens = function() {
    var res = this._newDamageRequestTokens;
    this._newDamageRequestTokens = undefined;
    return res;
  }

  //Draw tokens gained from this round. Compresses "unknown" tokens appropriately.
  Sprite_Character.prototype.updateDamageTokens = function(tkUnsort, ignoreGray) {
    var sprs = this._damageSprites;

    //Shuffle token array to contain "Real" first, followed by a single "I" for imaginary (since you typically only gain 1).
    var tokens = [];
    var hasI = false;
    for (var i=0; i<tkUnsort.length; i++) {
      var tk = tkUnsort[i];
      if (tk.startsWith("R") || tk.startsWith("P") || tk.startsWith("S")) {
        if (tk.length == 1) {
          tokens.push(tk);
        } else if (tk.endsWith("i") && !ignoreGray) {
          hasI = true;
        }
      }
    }
    if (hasI) {
      tokens.push("I");
    }
    //Now, print them in reverse order.
    for (var i=0; i<3; i++) {
      var spr = sprs[i];
      spr.opacity = 255;
      spr.bitmap.clearRect(spr.bitmap.rect.x,spr.bitmap.rect.y,spr.bitmap.rect.width,spr.bitmap.rect.height);

      var cx = spr.bitmap.width/2;
      var cy = spr.bitmap.height/2;
      var sz = cx < cy ? cx : cy;
      if (sz > 8) { sz = 8; }
      var tkId = i;
      if (tkId < tokens.length) {
        var token = tokens[tkId];
        var clr = "#999999";
        if (token === "R") {
          clr = $RageColor;
        } else if (token === "P") {
          clr = $PoiseColor;
        } else if (token === "S") {
          clr = $SerenityColor;
        }
        spr.bitmap.drawCircle(cx, cy, sz, "#000000");
        spr.bitmap.drawCircle(cx, cy, sz-1, "#FFFFFF");
        spr.bitmap.drawCircle(cx, cy, sz-2, clr);
      }
    }

    //Set offset from center of character.
    amt = 18;
    if (tokens.length == 3) {
      sprs[1]._dmgOffsetX = 0;
      sprs[0]._dmgOffsetX = sprs[1]._dmgOffsetX - amt;
      sprs[2]._dmgOffsetX = sprs[1]._dmgOffsetX + amt;
    } else if (tokens.length == 2) {
      sprs[1]._dmgOffsetX = sprs[1].width/2;
      sprs[0]._dmgOffsetX = sprs[1]._dmgOffsetX - amt;
    } else {
      sprs[0]._dmgOffsetX = 0;
    }
  }
  

  //Draw a damage value (green if <0) and set to visible (or invisible if nil)
  Sprite_Character.prototype.updateDamage = function(amt, isInjury) {
    var sprs = this._damageSprites;

    //Invis?
    if (amt == null) {
      for (var i=0; i<3; i++) {
        sprs[i].opacity = 0;
      }
    }

    //Bound.
    if (amt > 999) {
      amt = 999;
    }

    var nw = $numberBitmap.width/10;
    var nh = $numberBitmap.height/4;
    var sY = 0;
    if (isInjury) {
      sY = 3*nh;
    }
    if (amt < 0) {
      amt *= -1;
      sY = 2 * nh;
    }

    //Visibility, print
    var digits = (""+amt).split(""); //Digits array
    var charWidths = [0,0,0];
    for (var i=0; i<3; i++) {
      var spr = sprs[2-i];
      spr.opacity = 255;
      spr.bitmap.clearRect(spr.bitmap.rect.x,spr.bitmap.rect.y,spr.bitmap.rect.width,spr.bitmap.rect.height);

      if (amt >= 0) {
        //Anything?
        var digit = digits.length - i - 1;
        if (digit >= 0) {
          digit = digits[digit];
          spr.bitmap.blt($numberBitmap, digit*nw, sY, nw, nh, 0, 0);
          charWidths[2-i] = this.plungeCharaWidth(digit);
        }
      }
    }

    //Set offset from center of character.
    amt = 0;
    if (digits.length == 3) {
      sprs[1]._dmgOffsetX = 0;
      sprs[0]._dmgOffsetX = sprs[1]._dmgOffsetX - charWidths[0] - amt;
      sprs[2]._dmgOffsetX = sprs[1]._dmgOffsetX + charWidths[1] + amt;
    } else if (digits.length == 2) {
      sprs[2]._dmgOffsetX = charWidths[2]/2;
      sprs[1]._dmgOffsetX = sprs[2]._dmgOffsetX - charWidths[1] - amt;
    } else {
      sprs[2]._dmgOffsetX = 0;
    }
    
  }


  Sprite_Character.prototype.plungeCharaWidth = function(digit) {
    if (digit === "0") {
      return 21;
    } else if (digit === "1") {
      return 18;
    } else if (digit === "2") {
      return 24;
    } else if (digit === "3") {
      return 22;
    } else if (digit === "4") {
      return 24;
    } else if (digit === "5") {
      return 22;
    } else if (digit === "6") {
      return 24;
    } else if (digit === "7") {
      return 22;
    } else if (digit === "8") {
      return 22;
    } else if (digit === "9") {
      return 26;
    }
    return $numberBitmap.width/10 - 9; //Just guess
  }

  var _Sprite_Character_update = Sprite_Character.prototype.update;
  Sprite_Character.prototype.update = function() {
    _Sprite_Character_update.call(this);

    //Set up numbers?
    var newReq = this._character.getNewDamageRequest();
    var newInjury = this._character.getNewDamageInjury();
    var newReqTokens = this._character.getNewDamageRequestTokens();
    if (typeof(newReqTokens) !== "undefined") {
      this.updateDamageTokens(newReqTokens, this._character._plungeId==1);
    } else if (typeof(newReq) !== "undefined") {
      this.updateDamage(newReq, newInjury);
    }

    if (this._character._turnOnLifeCycle) {
      if (this._lifeCycle.opacity === 0) {
        this.turnOnLifeCycle();
      }
      this._character._turnOnLifeCycle = false;
    }

    if (this._character._turnOffLifeCycle) {
      if (this._lifeCycle.opacity !== 0) {
        this._lifeCycle.opacity = 0;
      }
      this._character._turnOffLifeCycle = false;
    }

    //Keep numbers in line with charas. We can't make them children directly, due to x_scale
    if (typeof(this._damageSprites) !== "undefined") {
      for (var i=0; i<3; i++) {
        var spr = this._damageSprites[i];
        spr.x = this.x - 16 + spr._dmgOffsetX;
        spr.y = this.y - 92 + this._character._dmgOffsetY;
      }
    }
  }
        

  //Update when flipH is set.
  var _Sprite_Character_isImageChanged = Sprite_Character.prototype.isImageChanged;
  Sprite_Character.prototype.isImageChanged = function() {
    if (_Sprite_Character_isImageChanged.call(this)) {
      return true;
    }

    if (this.flipH !== this._character.flipH) {
      this.flipH = this._character.flipH;
      return true;
    }

    return false;
  };


  //Properly update the sprite if we're in a plunge.
  var _Sprite_Character_updateCharacterFrame = Sprite_Character.prototype.updateCharacterFrame;
  Sprite_Character.prototype.updateCharacterFrame = function() {
    if (typeof(this._character.plungePattern()) !== "undefined") {
      // Get the correct 3x6 group
      var pw = this.bitmap.width/3;
      var pw2 = pw / 3;
      var ph = this.bitmap.height/6;
      var sx = 0;
      var sy = 0;
      if (this._character.plungePattern()[0] === "Idle") {
        sx = 0; sy = 0; //Idle animation
      } else if (this._character.plungePattern()[0] === "Physical") {
        sx = 0; sy = 1; //Ready/Physical animation
      } else if (this._character.plungePattern()[0] === "Swing") {
        sx = 1; sy = 1; //Swinging animation. NOTE: We actually take frame 0 from "Stabbing" and edit the feet to make it look like walking.
      } else if (this._character.plungePattern()[0] === "Item") {
        sx = 1; sy = 5; //Use Item animation
      } else if (this._character.plungePattern()[0] === "Evade") {
        sx = 0; sy = 5; //Evade animation
      } else if (this._character.plungePattern()[0] === "Escape") {
        sx = 2; sy = 0; //Escape animation
      } else if (this._character.plungePattern()[0] === "ReadyMagical") {
        sx = 0; sy = 2; //Ready/Magical animation
      } else if (this._character.plungePattern()[0] === "Crisis") {
        sx = 2; sy = 2; //Crisis animation
      } else if (this._character.plungePattern()[0] === "Dead") {
        sx = 2; sy = 5; //Dead animation
      }  //TODO: more

      //Now, get the correct frame.
      var patId = this._character.plungePattern()[1];
      if (patId >= 3) {
        patId = 1;
      }
      sx = sx * pw + patId * pw2;
      sy = sy * ph;

      //Save it.
      //TODO: Bush depth?
      this.setFrame(sx, sy, pw2, ph);
    } else {
      //Default
      _Sprite_Character_updateCharacterFrame.call(this);
    }
  };

})();
